//! This module provides a simple interface to exchange messages (`message::AvariceMessage`) between
//! Avarice and ue-server. The model is simple - we create an mpsc-channel of `Sender` and
//! `Receiver`, then pass `Sender` to the `network` sub-module and wait for its messages about
//! ue-servers' connections using `Receiver`.
use std::collections::HashMap;
use std::error::Error;
use std::io::Write;
use std::net::{SocketAddr, TcpStream};
use std::sync::mpsc::{channel, Receiver};
pub use writer::MessageWriter;

mod message;
mod network;
mod reader;
mod writer;
pub use message::AvariceMessage;
pub use network::{run_server, NetworkMessage};

/// For collecting messages from all connected ue-servers and providing a way to reply back.
pub struct AvariceServer {
    connected_links: HashMap<SocketAddr, Link>,
    receiver: Receiver<NetworkMessage>,
}

/// For representing a link to one of the connected ue-servers, can be used to send messages back.
/// To receive messages use `AvariceServer`'s `next()` method instead.
pub struct Link {
    ue_server_address: SocketAddr,
    writer: MessageWriter,
    writing_stream: TcpStream,
}

impl AvariceServer {
    /// Blocks until a new message arrives from one of the ue-servers. Returns a pair of `Link`,
    /// corresponding to the ue-server that sent next message and `AvariceMessage`,
    /// representing that message.
    pub fn next(&mut self) -> Option<(&mut Link, AvariceMessage)> {
        loop {
            match self.receiver.recv() {
                Ok(NetworkMessage::ConnectionEstablished(ue_server_address, writing_stream)) => {
                    //  If `ue_server_address` was already present in `self.connected_links`
                    //  hash map, then it means we have failed to clean it up after it
                    //  has disconnected. We can just throw away the old value here.
                    self.connected_links.insert(
                        ue_server_address,
                        Link {
                            ue_server_address,
                            writing_stream,
                            writer: MessageWriter::new(),
                        },
                    );
                    continue;
                }
                Ok(NetworkMessage::ConnectionLost(ue_server_address)) => {
                    self.connected_links.remove(&ue_server_address);
                    continue;
                }
                Ok(NetworkMessage::InvalidDataReceived(ue_server_address)) => {
                    self.connected_links.remove(&ue_server_address);
                    continue;
                }
                Ok(NetworkMessage::UEReceivedUpdate(ue_server_address, ue_received_bytes)) => {
                    if let Some(link) = self.connected_links.get_mut(&ue_server_address) {
                        link.update_ue_received_bytes(ue_received_bytes)
                    }
                    continue;
                }
                Ok(NetworkMessage::MessageReceived(ue_server_address, message)) => {
                    //  Not having a link with key `ue_server_address` should be impossible here
                    return self
                        .connected_links
                        .get_mut(&ue_server_address)
                        .and_then(|x| Some((x, message)));
                }
                _ => return None,
            }
        }
    }
}

impl Link {
    pub fn send(&mut self, message: AvariceMessage) {
        self.writer.push(&message.to_string());
        self.flush();
    }

    pub fn socket_address(&self) -> SocketAddr {
        self.ue_server_address
    }

    fn update_ue_received_bytes(&mut self, ue_received_bytes: u64) {
        self.writer.update_ue_received_bytes(ue_received_bytes);
        self.flush();
    }

    fn flush(&mut self) {
        if let Some(bytes) = self.writer.try_pop() {
            self.writing_stream.write_all(&bytes).unwrap();
        }
    }
}

/// Creates a new `AvariceServer` that will listen for ue-server connections on the specified port.
pub fn start_avarice(port: u16) -> Result<AvariceServer, Box<dyn Error>> {
    let (sender, receiver) = channel();
    run_server(port, sender)?;
    Ok(AvariceServer {
        connected_links: HashMap::new(),
        receiver,
    })
}
