//!     Implements a network model where messages from all the ue-servers are collected in a single
//! main thread. For that we spawn a new thread that listens for new connections, which in turn
//! spawns a new thread for every connected ue-server to handle reading data from it.
//!     Since all reading is handled in ue-servers' own threads, to collect messages they have
//! received in the main thread we use `std::sync::mpsc::Sender`. Conversely, all the writing to
//! ue-server is handled in the main thread itself. Writing `TcpStream` is sent to the main thread
//! by the same `std::sync::mpsc::Sender` object.
use super::message::AvariceMessage;
pub use super::reader::MessageReader;
use std::error::Error;
use std::io::Read;
use std::net::{SocketAddr, TcpListener, TcpStream};
use std::sync::mpsc::Sender;
use std::thread;

pub struct UEConnection {
    pub address: SocketAddr,
    pub reader: MessageReader,
    pub reading_stream: TcpStream,
    pub message_sender: Sender<NetworkMessage>,
}

/// Possible messages to the main thread
pub enum NetworkMessage {
    ConnectionEstablished(SocketAddr, TcpStream),
    InvalidDataReceived(SocketAddr),
    ConnectionLost(SocketAddr),
    MessageReceived(SocketAddr, AvariceMessage),
    UEReceivedUpdate(SocketAddr, u64),
}

pub fn run_server(port: u16, message_sender: Sender<NetworkMessage>) -> Result<(), Box<dyn Error>> {
    let address = SocketAddr::from(([0, 0, 0, 0], port));
    let listener = TcpListener::bind(address)?;
    thread::spawn(move || loop {
        //  Listen to new (multiple) connection
        let (reading_stream, address) = listener.accept().unwrap();
        let writing_stream = reading_stream.try_clone().unwrap();
        message_sender
            .send(NetworkMessage::ConnectionEstablished(
                address,
                writing_stream,
            ))
            .unwrap();
        //  On connection - spawn a new thread
        let sender_clone = message_sender.clone();
        thread::spawn(move || {
            manage_connection(UEConnection {
                reader: MessageReader::new(),
                message_sender: sender_clone,
                reading_stream,
                address,
            })
        });
    });
    Ok(())
}

fn manage_connection(mut connection: UEConnection) {
    let mut buffer = [0; 1024];
    loop {
        //  Reading cycle
        match connection.reading_stream.read(&mut buffer) {
            Ok(n) => connection.reader.push(&buffer[..n]).unwrap(),
            _ => {
                connection
                    .message_sender
                    .send(NetworkMessage::ConnectionLost(connection.address))
                    .unwrap();
                return;
            }
        };
        if connection.reader.is_broken() {
            connection
                .message_sender
                .send(NetworkMessage::InvalidDataReceived(connection.address))
                .unwrap();
            return;
        }
        //  Decoding cycle
        while let Some(text_message) = connection.reader.pop() {
            if let Some(avarice_message) = AvariceMessage::from(&text_message) {
                connection
                    .message_sender
                    .send(NetworkMessage::MessageReceived(
                        connection.address,
                        avarice_message,
                    ))
                    .unwrap();
            } else {
                connection
                    .message_sender
                    .send(NetworkMessage::InvalidDataReceived(connection.address))
                    .unwrap();
                return;
            }
        }
        connection
            .message_sender
            .send(NetworkMessage::UEReceivedUpdate(
                connection.address,
                connection.reader.ue_received_bytes(),
            ))
            .unwrap();
    }
}
